/**
 * Smart Waste Sorting System - Final Project
 *
 * Group Members:
 * Abbas, Mustafa, Aina, Zainab, Asma, Saif
 *
 * Description:
 * This Arduino UNO R4 WiFi code detects and sorts different types of waste
 * using sensors (ultrasonic, IR, LDR) and servo motors. It also connects
 * to WiFi and serves sensor data via a web interface.
 */

#include <WiFiS3.h>        /** WiFi library for UNO R4 WiFi **/
#include <ArduinoJson.h>   /** For generating JSON responses **/
#include <Servo.h>         /** For controlling servo motors **/

// WiFi credentials
const char* ssid = "AS.12";               /** WiFi SSID (network name) **/
const char* password = "asmalool";        /** WiFi password **/

// Web server on port 80
WiFiServer server(80);                    /** Create a web server on port 80 **/

// Sensor pins
const int metalSensorPin = 5;             /** Ultrasonic sensor for metal bin **/
const int organicSensorPin = 6;           /** Ultrasonic sensor for organic bin **/
const int glassSensorPin = 7;             /** Ultrasonic sensor for glass bin **/
const int irPin = 4;                      /** IR sensor pin (detect presence) **/
const int ldrPin = A0;                    /** LDR analog pin (light detection) **/
const int ledPin = 3;                     /** System status LED pin **/
const int debugLed = 13;                  /** Debug LED pin **/
const int redLed = 2;                     /** Red alert LED pin **/
const int servo1Pin = A1;                 /** Servo motor 1 control pin **/
const int servo2Pin = A2;                 /** Servo motor 2 control pin **/

// Servo setup
Servo servo1, servo2;                     /** Create two servo motor objects **/
int servo1Angle = 0;                      /** Initial angle for servo1 **/
int servo2Angle = 180;                    /** Initial angle for servo2 **/

// Sensor values
float metalDistance = 0.0;                /** Distance from metal sensor **/
float organicDistance = 0.0;              /** Distance from organic sensor **/
float glassDistance = 0.0;                /** Distance from glass sensor **/
float primaryDistance = 0.0;              /** Placeholder for primary reading **/
int irValue = 0;                          /** IR sensor reading **/
int ldrValue = 0;                         /** LDR light intensity reading **/
unsigned long lastSensorUpdate = 0;       /** Last time sensors were updated **/
const unsigned long sensorUpdateInterval = 1000; /** Time interval for updating sensors **/

// Rotation control
bool rotating = false;                    /** Is a servo currently rotating? **/
int rotationStepsRemaining = 0;           /** Remaining steps for rotation **/
int rotationDirection = 1;                /** Direction of rotation: 1 = CW, -1 = CCW **/
Servo* activeServo = nullptr;             /** Pointer to currently active servo **/

// --- Setup ---
void setup() {
  Serial.begin(9600);                     /** Start serial monitor for debugging **/

  pinMode(irPin, INPUT);                  /** Set IR pin as input **/
  pinMode(ldrPin, INPUT);                 /** Set LDR pin as input **/
  pinMode(ledPin, OUTPUT);                /** Set LED pin as output **/
  pinMode(debugLed, OUTPUT);              /** Set debug LED as output **/
  pinMode(redLed, OUTPUT);                /** Set red LED as output **/
  digitalWrite(ledPin, HIGH);             /** Turn on status LED **/
  digitalWrite(redLed, HIGH);             /** Turn on red LED **/

  servo1.attach(servo1Pin);               /** Attach servo1 to its pin **/
  servo2.attach(servo2Pin);               /** Attach servo2 to its pin **/

  connectToWiFi();                        /** Connect to WiFi network **/
  server.begin();                         /** Start the HTTP server **/

  Serial.println("Web server started!");
  Serial.print("Access sensor data at: http://");
  Serial.println(WiFi.localIP());         /** Print device IP address **/

  updateSensorData();                     /** Get initial sensor readings **/
}

// --- Loop ---
void loop() {
  static unsigned long lastWiFiCheck = 0;

  if (millis() - lastWiFiCheck >= 30000) {
    checkWiFiStatus();                    /** Recheck WiFi every 30s **/
    lastWiFiCheck = millis();
  }

  if (millis() - lastSensorUpdate >= sensorUpdateInterval) {
    updateSensorData();                   /** Refresh sensor readings **/
    lastSensorUpdate = millis();
  }

  irValue = digitalRead(irPin);           /** Read IR sensor **/
  ldrValue = analogRead(ldrPin);          /** Read LDR sensor **/
  Serial.print("LDR: "); Serial.println(ldrValue);
  digitalWrite(debugLed, irValue == 0 ? HIGH : LOW); /** Turn on debug LED if object detected **/

  handleLDRAndRotateServo();              /** Trigger servo based on light value **/

  WiFiClient client = server.available(); /** Listen for incoming web client **/
  if (client) {
    Serial.println("New client connected");
    handleClient(client);                 /** Handle web request **/
  }
}

// --- WiFi ---
void connectToWiFi() {
  if (WiFi.status() == WL_NO_MODULE) {
    Serial.println("No WiFi module detected!");
    while (true);                         /** Halt if WiFi module is missing **/
  }

  WiFi.disconnect();                      /** Reset WiFi connection **/
  delay(1000);
  WiFi.begin(ssid, password);             /** Start WiFi connection **/

  int attempts = 0;
  while ((WiFi.status() != WL_CONNECTED || WiFi.localIP() == IPAddress(0, 0, 0, 0)) && attempts < 60) {
    delay(1000);
    Serial.print("Attempt ");
    Serial.print(attempts + 1);
    Serial.print(" - Status: ");
    Serial.println(getWiFiStatusText(WiFi.status()));
    attempts++;
  }

  if (WiFi.status() == WL_CONNECTED && WiFi.localIP() != IPAddress(0, 0, 0, 0)) {
    Serial.println("✅ Connected!");
    Serial.print("📡 IP Address: ");
    Serial.println(WiFi.localIP());
  } else {
    Serial.println("❌ Failed to get a valid IP address!");
    scanNetworks();                       /** Try scanning available networks **/
  }
}

void checkWiFiStatus() {
  if (WiFi.status() != WL_CONNECTED || WiFi.localIP() == IPAddress(0, 0, 0, 0)) {
    Serial.println("WiFi lost or IP invalid. Reconnecting...");
    connectToWiFi();                      /** Reconnect to WiFi if lost **/
  }
}

void scanNetworks() {
  int n = WiFi.scanNetworks();            /** Scan for nearby WiFi networks **/
  for (int i = 0; i < n; i++) {
    Serial.print(i + 1);
    Serial.print(": ");
    Serial.print(WiFi.SSID(i));
    Serial.print(" (");
    Serial.print(WiFi.RSSI(i));
    Serial.println(" dBm)");
  }
}

String getWiFiStatusText(int status) {
  switch (status) {
    case WL_IDLE_STATUS: return "Idle";
    case WL_NO_SSID_AVAIL: return "No SSID Available";
    case WL_SCAN_COMPLETED: return "Scan Completed";
    case WL_CONNECTED: return "Connected";
    case WL_CONNECT_FAILED: return "Connection Failed";
    case WL_CONNECTION_LOST: return "Connection Lost";
    case WL_DISCONNECTED: return "Disconnected";
    case WL_NO_MODULE: return "No WiFi Module";
    default: return "Unknown";
  }
}

// --- Sensor ---
void updateSensorData() {
  metalDistance = getDistance(metalSensorPin);     /** Get metal bin distance **/
  organicDistance = getDistance(organicSensorPin); /** Get organic bin distance **/
  glassDistance = getDistance(glassSensorPin);     /** Get glass bin distance **/
  primaryDistance = metalDistance;                 /** Use metal as default **/
}

float getDistance(int pin) {
  pinMode(pin, OUTPUT);
  digitalWrite(pin, LOW); delayMicroseconds(2);
  digitalWrite(pin, HIGH); delayMicroseconds(10);
  digitalWrite(pin, LOW);
  pinMode(pin, INPUT);
  long duration = pulseIn(pin, HIGH, 30000);        /** Wait for echo signal **/
  return (duration == 0) ? -1 : duration * 0.034 / 2;/** Convert to cm **/
}

// --- Rotation Logic ---
void rotateServo(Servo& servo, int rotations, bool forward = true) {
  rotating = true;
  rotationStepsRemaining = 360 * rotations;
  rotationDirection = forward ? 1 : -1;
  activeServo = &servo;

  int currentAngle = servo.read();
  while (rotationStepsRemaining > 0) {
    currentAngle += rotationDirection;
    if (currentAngle > 180) currentAngle = 0;
    else if (currentAngle < 0) currentAngle = 180;

    servo.write(currentAngle);
    delay(10); // speed
    rotationStepsRemaining--;
  }

  rotating = false;
  activeServo = nullptr;
}

void handleLDRAndRotateServo() {
  if (rotating) return;

  if (ldrValue < 200) {
    Serial.println("✅ Glass detected");
    rotateServo(servo1, 2, true); // glass
  }
  else if (ldrValue >= 200 && ldrValue <= 220) {
    Serial.println("⏸️ No action (neutral range)");
  }
  else if (ldrValue > 220 && ldrValue < 260) {
    Serial.println("🛠️ Metal detected");
    rotateServo(servo1, 4, true); // metal
  }
  else {
    Serial.println("🍌 Organic detected");
    rotateServo(servo2, 2, false); // organic
  }
}

// --- Web Server Handlers ---
void handleClient(WiFiClient client) {
  String currentLine = "", request = "";
  while (client.connected()) {
    if (client.available()) {
      char c = client.read();
      if (c == '\n') {
        if (currentLine.length() == 0) {
          processRequest(client, request);
          break;
        } else {
          if (request == "") request = currentLine;
          currentLine = "";
        }
      } else if (c != '\r') {
        currentLine += c;
      }
    }
  }
  client.stop();
}

void processRequest(WiFiClient client, String request) {
  if (request.startsWith("GET /")) {
    String path = extractPath(request);
    if (path == "/" || path == "/status") sendStatusResponse(client);
    else if (path == "/distance/json") sendJSONResponse(client);
    else send404Response(client);
  } else {
    sendMethodNotAllowedResponse(client);
  }
}

String extractPath(String request) {
  int start = request.indexOf(' ') + 1;
  int end = request.indexOf(' ', start);
  return request.substring(start, end);
}

// --- Responses ---
void sendStatusResponse(WiFiClient client) {
  String html = "<!DOCTYPE html><html><head><title>Status</title><meta http-equiv='refresh' content='2'>";
  html += "<style>body{font-family:Arial;background:#111;color:#fff;padding:20px;}</style></head><body>";
  html += "<h1>Sensor Monitor</h1>";
  html += "<p><strong>Metal:</strong> " + String(metalDistance, 1) + " cm</p>";
  html += "<p><strong>Organic:</strong> " + String(organicDistance, 1) + " cm</p>";
  html += "<p><strong>Glass:</strong> " + String(glassDistance, 1) + " cm</p>";
  html += "<p><strong>IR:</strong> " + String(irValue) + "</p>";
  html += "<p><strong>LDR:</strong> " + String(ldrValue) + "</p>";
  html += "</body></html>";
  sendHTTPResponse(client, "200 OK", "text/html", html);
}

void sendJSONResponse(WiFiClient client) {
  JsonDocument doc;
  doc["metal_cm"] = metalDistance;
  doc["organic_cm"] = organicDistance;
  doc["glass_cm"] = glassDistance;
  doc["ir"] = irValue;
  doc["ldr"] = ldrValue;
  doc["timestamp"] = millis();
  String json;
  serializeJson(doc, json);
  sendHTTPResponse(client, "200 OK", "application/json", json);
}

void send404Response(WiFiClient client) {
  sendHTTPResponse(client, "404 Not Found", "application/json", "{\"error\":\"Endpoint not found\"}");
}

void sendMethodNotAllowedResponse(WiFiClient client) {
  sendHTTPResponse(client, "405 Method Not Allowed", "application/json", "{\"error\":\"Method not allowed\"}");
}

void sendHTTPResponse(WiFiClient client, String status, String contentType, String body) {
  client.println("HTTP/1.1 " + status);
  client.println("Content-Type: " + contentType);
  client.println("Content-Length: " + String(body.length()));
  client.println("Connection: close");
  client.println("Access-Control-Allow-Origin: *");
  client.println();
  client.print(body);
}